#include <stdio.h>
#include <stdlib.h>

#include "gmp.h"
#include "gmp-impl.h"

/* Computes a^e (mod B). Uses right-to-left binary algorithm, since
   typical use will have e small. */
static mp_limb_t
powlimb (mp_limb_t a, mp_limb_t e)
{
  mp_limb_t r = 1;
  mp_limb_t s = a;

  for (r = 1, s = a; e > 0; e >>= 1, s *= s)
    if (e & 1)
      r *= s;

  return r;
}

/* Computes the nth root of A, mod B^k. Both A and n must be odd.

   Uses the iteration

     x' <-- x - x * (a^{n-1} x^n - 1) / n

   converging to a^{1/n - 1}.

   If

     a^{n-1} x^n = 1 (mod 2^\ell),

   then

     a^{n-1} x'^n = 1 (mod 2^{2\ell}),
*/

void
mpn_broot (mp_ptr rp, mp_srcptr ap, mp_size_t an, mp_limb_t n)
{
  mp_size_t sizes[GMP_LIMB_BITS * 2];  
  mp_ptr anm1, tp, xp, xnp, ep;
  mp_limb_t a0, r0, nm1, ninv;
  mp_size_t xn;
  unsigned i;

  TMP_DECL;

  ASSERT (an > 0);
  ASSERT (ap[0] & 1);
  ASSERT (n & 1);
  ASSERT (n >= 3);

  TMP_MARK;
  
  anm1 = TMP_ALLOC_LIMBS (4*an);
  tp = anm1 + an;

  nm1 = n-1;
  mpn_powlo (anm1, ap, &nm1, 1, an, tp); /* 3 an scratch space */

  a0 = ap[0];
  binvert_limb (ninv, n);

  /* 4 bits: a^{1/n - 1} (mod 16):

       a mod 8
       1 3 5 7   
     1 1 1 1 1
     3 1 9 9 1
  */
  r0 = 1 + (((n << 2) & ((a0 << 1) ^ (a0 << 2))) & 8);
  r0 = ninv * r0 * (n+1 - anm1[0] * powlimb (r0, n)); /* 8 bits */
  r0 = ninv * r0 * (n+1 - anm1[0] * powlimb (r0, n)); /* 16 bits */
  r0 = ninv * r0 * (n+1 - anm1[0] * powlimb (r0, n)); /* 32 bits */
#if GMP_NUMB_BITS > 32
  r0 = ninv * r0 * (n+1 - anm1[0] * powlimb (r0, n)); /* 64 bits */
#endif

  if (an == 1)
    {
      TMP_FREE;
      rp[0] = r0 * a0;
      return;
    }

  xp = TMP_ALLOC_LIMBS (3*an);
  xnp = xp + an;
  ep = xp + 2*an;

  /* FIXME: Possible to this on the fly with some bit fiddling. */
  for (i = 0; an > 1; an = (an + 1)/2)
    sizes[i++] = an;

  xp[0] = r0;
  xn = 1;

  while (i-- > 0)
    {
      /* Compute x^n. What's the best way to handle the doubled
	 precision? Could do the complete powering using
	 wraparound. */
      MPN_ZERO (xp + xn, sizes[i] - xn);
      mpn_powlo (xnp, xp, &n, 1, sizes[i], tp); 

      /* Multiply by a^{n-1}. Can use wraparound; low part is
	 000...01. */

      mpn_mullo_n (ep, xnp, anm1, sizes[i]);
      ASSERT (ep[0] == 1);
      ASSERT (xn == 1 || mpn_zero_p (ep + 1, xn - 1));

      ASSERT (sizes[i] <= 2*xn);
      mpn_pi1_bdiv_q_1 (ep, ep + xn, sizes[i] - xn, n, ninv, 0);

      /* Multiply by x, plain mullo. */
      mpn_mullo_n (xp + xn, ep, xp, sizes[i] - xn);

      /* FIXME: Avoid negation, e.g., by using a bdiv_q_1 variant
	 returning -q. */
      mpn_neg (xp + xn, xp + xn, sizes[i] - xn);

      xn = sizes[i];
    }
  mpn_mullo_n (rp, ap, xp, xn);
  TMP_FREE;
}

#define MAX_LIMBS 150

int
main (int argc, char **argv)
{
  mp_limb_t a[MAX_LIMBS];
  mp_limb_t r[MAX_LIMBS];
  mp_limb_t t[4*MAX_LIMBS];

  gmp_randstate_t rands;
  unsigned i;
  
  gmp_randinit_default (rands);

  for (i = 0; i < 5000; i++)
    {
      mp_size_t s;
      mp_limb_t n;
      int c;

      s = 1 + gmp_urandomm_ui (rands, MAX_LIMBS);

      if (i & 1)
	mpn_random2 (a, s);
      else
	mpn_random (a, s);

      a[0] |= 1;

      if (i < 100)
	n = 3 + 2*i;
      else
	{
	  mpn_random (&n, 1);
	  if (n < 3)
	    n = 3;
	  else
	    n |= 1;
	}

      mpn_broot (r, a, s, n);
      mpn_powlo (t, r, &n, 1, s, t + s);

      MPN_CMP (c, t, a, s);
      if (c != 0)
	{
	  gmp_fprintf (stderr,
		       "mpn_broot returned bad result: %d limbs\n",
		       (int) s);
	  gmp_fprintf (stderr, "n   = %Mx\n", n, s);
	  gmp_fprintf (stderr, "a   = %Nx\n", a, s);
	  gmp_fprintf (stderr, "r   = %Nx\n", r, s);
	  gmp_fprintf (stderr, "r^n = %Nx\n", t, s);
	  abort ();
	}
    }
  return EXIT_SUCCESS;
}
